;;; czech-calendar.el -- Czech calendar and holidays

;; Copyright (C) 1997, 1998 Milan Zamazal

;; Authors:    David auer <xsauer@hwlab.felk.cvut.cz>,
;;             Milan Zamazal <pdm@freesoft.cz>
;; Maintainer: Milan Zamazal <pdm@freesoft.cz>
;; Version:    4.3
;; Keywords:   i18n, Czech, calendar, holidays

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 2 of the License.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You can find the GNU General Public License at
;; http://www.gnu.org/copyleft/gpl.html
;; or you can write to the Free Software Foundation, Inc., 675 Mass Ave.,
;; Cambridge, MA 02139, USA.

;;; Commentary:

;; This file defines and redefines various calendar things according to Czech
;; and Moravian national habits.

;;; History:

;; So long, so very long...

;;; Code:


;; Holidays with names
(defvar czech-calendar-names nil
  "*Whether Czech calendar should contain all calendar names.")

;; Basic calendar settings according to Czech habits
(setq european-calendar-style t
      calendar-week-start-day 1
      calendar-day-name-array
      ["Nedle" "Pondl" "ter" "Steda" "tvrtek" "Ptek" "Sobota"]
      calendar-month-name-array
      ["Leden" "nor" "Bezen" "Duben" "Kvten" "erven"
       "ervenec" "Srpen" "Z" "jen" "Listopad" "Prosinec"]
      display-time-24hr-format t
      cal-tex-24 t
      cal-tex-options "czech"
      calendar-date-display-form '(day ". " month ". " year)
      calendar-time-display-form '(24-hours ":" minutes
					    (if time-zone " (")
					    time-zone
					    (if time-zone ")")))

;; Do not allow short month names
(defun czech-calendar-hack ()
  (defun calendar-month-name (month &optional width)
    "The name of MONTH.
Redefinition of the original function because of some difficulties with Czech.
WIDTH is ignored because of this reason."
    (aref calendar-month-name-array (1- month))))
(if (featurep 'calendar)
    (czech-calendar-hack))
(add-hook 'calendar-load-hook 'czech-calendar-hack)

;; Czech holidays
(setq local-holidays
      '((holiday-fixed
	 1 1    "Nov rok")
	(holiday-fixed
	 2 14   "Svtek sv. Valentna")
	(holiday-fixed
	 4 1    "Aprl!")
	(holiday-fixed
	 5 1    "Svtek prce")
	(holiday-float
	 5 0 2  "Den matek")
	(holiday-fixed
	 5 5    "Prask povstn (1945)")
	(holiday-fixed
	 5 8    "Den osvobozen od faismu (1945)")
	(holiday-fixed
	 7 5    "Den slovanskch vrozvst Cyrila a Metodje")
	(holiday-fixed
	 7 6    "Mistr Jan Hus (1415)")
	(holiday-fixed
	 10 28  "Den vzniku samostatnho eskoslovenskho sttu (1918)")
	(holiday-fixed
	 11 2   "Pamtka zesnulch")
	(holiday-fixed
	 11 17  "Den boje student za svobodu a demokracii")
	(holiday-fixed
	 12 24  "tdr den")
	(holiday-fixed
	 12 26  "2. svtek vnon")
	(holiday-fixed
	 12 31  "Silvestr")))
(when czech-calendar-names
  (load "czech-calendar-names"))

;; Translate daylight savings time
(setq solar-holidays
  '((solar-equinoxes-solstices)
    (funcall
     'holiday-sexp
     calendar-daylight-savings-starts
     '(format "Zan letn as"))
    (funcall
     'holiday-sexp
     calendar-daylight-savings-ends
     '(format "Kon letn as"))))

;; Translate solar seasons
(setq solar-n-hemi-seasons
  '("Jarn rovnodennost" "Letn slunovrat" 
    "Podzimn rovnodennost" "Zimn slunovrat"))

;; Translate Christian holidays
(setq christian-holidays
  '((if all-christian-calendar-holidays
        (holiday-fixed 1 6 "Ti krlov"))
    (cz-holiday-easter-etc)
    (if all-christian-calendar-holidays
        (holiday-greek-orthodox-easter))
    (if all-christian-calendar-holidays
        (holiday-fixed 8 15 "Nanebvzet Panny Marie"))
    (if all-christian-calendar-holidays
        (holiday-advent))
    (holiday-fixed 12 25 "1. svtek vnon")
    (if all-christian-calendar-holidays
        (holiday-julian 12 25 "Pravoslavn Vnoce"))))

;; Another replacing case.  Cut, paste & edit from `holidays.el' again.
;; Here we moreover slightly modify produced list, because in Czech Republic
;; there is another set of mandatory Easter holidays.
(defun cz-holiday-easter-etc ()
  "List of dates related to Easter, as visible in calendar window."
  (if (and (> displayed-month 5) (not all-christian-calendar-holidays))
      nil;; Ash Wednesday, Good Friday, and Easter are not visible.
    (let* ((century (1+ (/ displayed-year 100)))
	   (shifted-epact;; Age of moon for April 5...
	    (% (+ 14 (* 11 (% displayed-year 19));;     ...by Nicaean rule
		  (-;; ...corrected for the Gregorian century rule
		   (/ (* 3 century) 4))
		  (/;; ...corrected for Metonic cycle inaccuracy.
		   (+ 5 (* 8 century)) 25)
		  (* 30 century));;              Keeps value positive.
	       30))
	   (adjusted-epact;;  Adjust for 29.5 day month.
	    (if (or (= shifted-epact 0)
		    (and (= shifted-epact 1) (< 10 (% displayed-year 19))))
		(1+ shifted-epact)
	      shifted-epact))
	   (paschal-moon;; Day after the full moon on or after March 21.
	    (- (calendar-absolute-from-gregorian
		(list 4 19 displayed-year))
	       adjusted-epact))
	   (abs-easter (calendar-dayname-on-or-before
			0 (+ paschal-moon 7)))
	   (mandatory
	    (list
	     (list (calendar-gregorian-from-absolute abs-easter)
		   "Velikonon nedle")
	     (list (calendar-gregorian-from-absolute (1+ abs-easter))
		   "Velikonon pondl")))
	   (optional
	    (list
	     (list (calendar-gregorian-from-absolute (- abs-easter 49))
		   "Masopustn nedle")
	     (list (calendar-gregorian-from-absolute (- abs-easter 48))
		   "Masopustn pondl")
	     (list (calendar-gregorian-from-absolute (- abs-easter 47))
		   "Masopustn ter")
	     (list (calendar-gregorian-from-absolute (- abs-easter 46))
		   "Popelen steda")
	     (list (calendar-gregorian-from-absolute (- abs-easter 14))
		   "Smrtn nedle")
	     (list (calendar-gregorian-from-absolute (- abs-easter 7))
		   "Kvtn nedle")
	     (list (calendar-gregorian-from-absolute (- abs-easter 3))
		   "Zelen tvrtek")
	     (list (calendar-gregorian-from-absolute (- abs-easter 2))
		   "Velk ptek")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 39))
		   "Nanebevstoupen Pn")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 49))
		   "Svatodun nedle")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 50))
		   "Svatodun pondl")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 56))
		   "Nejsvtj Trojice")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 60))
		   "Tlo a Krev Pn")
	     (list (calendar-gregorian-from-absolute (+ abs-easter 68))
		   "Nejsvtj Srdce Jeovo")))
	   (output-list
	    (filter-visible-calendar-holidays mandatory)))
      (if all-christian-calendar-holidays
	  (setq output-list
		(append 
		 (filter-visible-calendar-holidays optional)
		 output-list)))
      output-list)))


;;; *** Announce ***

(provide 'czech-calendar)


;; Local variables:
;; coding: iso-latin-2
;; End:

;;; czech-calendar.el ends here

