;;; czech-message.el -- Czech support for mail/news

;; Copyright (C) 1996-1998 Milan Zamazal

;; Author:     Milan Zamazal <pdm@fi.muni.cz>
;; Version:    4.3

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 2 of the License.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You can find the GNU General Public License at
;; http://www.gnu.org/copyleft/gpl.html
;; or you can write to the Free Software Foundation, Inc., 675 Mass Ave.,
;; Cambridge, MA 02139, USA.

;;; Commentary:

;; It supports Czech specific handling of mail and news.  Currently only Gnus
;; and RMAIL are supported.

;;; Installation:

;; Put the following line into your ~/.gnus.el (or ~/.emacs if you use RMAIL):
;;
;;   (require 'czech-message)

;;; Code:



(require 'message)
(require 'rfc822)


;;; User configuration variables

(defvar cz-gnus 'no-mime
  "*Whether emacs-czech should attack Gnus.
Possible values are the following:
`nil'     - no connection to Gnus
`no-mime' - enable Czech hooks for messages but do not use MIME (default)
`t'       - enable Czech hooks for messages and use `semi' MIME package")

(defvar cz-message-method-default 'ascii
  "*Default method used for mail/news conversion.
Its values can be the following symbols:
  `ascii' ... convert to 7-bit ASCII (default)
  `mime'  ... convert to MIME
  `plain' ... no conversion
  `nil'   ... the same as `plain'
If the value is `mime' and the value of `cz-gnus' is `no-mime', then this
option acts in the same way as `ascii'.
It can be overriden by variables `cz-message-method-mail' and
`cz-message-method-news'.")

(defvar cz-message-method-mail 'mime
  "*Default method used for mail conversion.
Its values can be the following symbols:
  `ascii' ... convert to 7-bit ASCII
  `mime'  ... convert to MIME (default)
  `plain' ... no conversion
  `nil'   ... use the value of variable `cz-message-method-default'
If the value is `mime' and the value of `cz-gnus' is `no-mime', then this
option acts in the same way as `ascii'.")

(defvar cz-message-method-news 'mime
  "*Default method used for news conversion.
Its values can be the following symbols:
  `ascii' ... convert to 7-bit ASCII
  `mime'  ... convert to MIME
  `plain' ... no conversion
  `nil'   ... use the value of variable `cz-message-method-default' (default)
If the value is `mime' and the value of `cz-gnus' is `no-mime', then this
option acts in the same way as `ascii'.")

(defvar cz-message-method-people '((mime "pdm@fi.muni.cz"))
  "*List of conversion methods for mail recipients.
This is a list of lists.
Each of sublists begins with one of the following symbols:
 `ascii' mails for listed people are converted to 7-bit ASCII
 `mime'  mails for listed people are converted to 7-bit MIME
 `plain' mails for listed people are not converted or are converted (if
         possible) to 8-bit MIME
Follow regular expressions of respective e-mail addresses.
See also variable `cz-message-method-mail'.")

(defvar cz-message-method-newsgroups nil
  "*List of conversion methods for newsgroups.
This is a list of lists.
Each of sublists begins with one of the following symbols:
 `ascii' listed articles are converted to 7-bit ASCII
 `mime'  listed articles are converted to MIME
 `plain' listed articles are not converted or are converted (if
         possible) to 8-bit MIME
Follow regular expressions of respective newsgroups.
See also variable `cz-message-method-news'.")

(defvar cz-message-use-bbdb nil
  "*If non-nil, use BBDB for selecting message conversion method.
You can use Insidious Big Brother Database for selecting message conversion
method by specifying field `mailtype'.")

(defvar cz-message-ask-on-send t
  "*Should I ask for confirmation after performing conversion on message?")

(defvar cz-message-evil-sysadmin nil
  "*Set this to t if your sendmail installation converts 8-bit mails to MIME.")

(defvar cz-message-using-semi nil
  "*Set this to t if you are using SEMI based MIME support.")

;;; Internal variables

(defvar cz-message-this-method 'default
  "Encoding of the current message.")
(make-variable-buffer-local 'cz-message-this-method)

(defvar cz-message-before-hook nil
  "Hooks run before Czech message is beeing prepared.")

(defvar cz-message-after-hook nil
  "Hooks run after Czech message has been prepared.")


;;; *** Main part ***

;;;###autoload
(defun cz-message-send (&optional arg)
  "Run `cz-message-prepare' and then `message-send'.
With an prefix argument send the message without Czech processing."
  (interactive "P")
  (if arg
      (if (cz-message-rmail-p)
	  (mail-send)
	(message-send))
    (run-hooks 'cz-message-before-hook)
    (if (cz-message-prepare)
	(progn
	  (run-hooks 'cz-message-after-hook)
	  (if (cz-message-rmail-p)
	      (mail-send)
	    (message-send)))
      (run-hooks 'cz-message-after-hook))))

;;;###autoload
(defun cz-message-send-and-exit (&optional arg)
  "Run `cz-message-prepare' and then `message-send-and-exit'.
With an prefix argument send the message without Czech processing"
  (interactive "P")
  (if arg
      (if (cz-message-rmail-p)
	  (mail-send-and-exit)
	(message-send-and-exit))
    (run-hooks 'cz-message-before-hook)
    (if (cz-message-prepare)
	(progn
	  (run-hooks 'cz-message-after-hook)
	  (if (cz-message-rmail-p)
	      (mail-send-and-exit)
	    (message-send-and-exit)))
      (run-hooks 'cz-message-after-hook))))

(defun cz-message-rmail-p ()
  "Return non-nil if we are in RMAIL mail composition buffer."
  (eq major-mode 'mail-mode))

(defun cz-message-prepare ()
  "Perform conversion on the current message buffer and ask for confirmation.
See variables `cz-gnus', `cz-message-method-default',
`cz-message-method-people', and `cz-message-news' for basic explanation, what
conversion will be performed."
  (let ((method (cz-message-show-method 'quietly)))
    ;; Handle methods
    (save-excursion
      (cond
       ((eq method 'ascii)
	(cz-message-convert-to-csech)
	(message "Message converted to 7-bit ASCII.")
	(cz-message-convert-to-mime 7))	; we may use MIME anyway
       ((eq method 'mime)
	(cz-message-convert-to-mime 7))
       ((or (eq method 'mime-8) (eq method 'plain))
	(cz-message-convert-to-mime 8))
       ((eq method 'digraphs)
	(cz-message-convert-to-digraphs))
       ((eq method nil)
	(message "Message NOT converted."))
       (t (error "Unknown conversion method: %s" (symbol-name method)))))
    ;; Ask for confirmation
    (or (not cz-message-ask-on-send)
	(yes-or-no-p
	 (format "Is the %s correct? " (if (cz-message-rmail-p)
					   "MAIL"
					 (if (message-mail-p)
					     (if (message-news-p)
					    "MAIL and ARTICLE"
					  "MAIL")
					"ARTICLE")))))))

(defun cz-message-convert-to-mime (bit-level)
  "Convert message to MIME.
BIT-LEVEL is either 7 or 8."
  (when cz-message-using-semi
    (mime-edit-toggle-transfer-level bit-level)
    (if mime-edit-mode-flag
	(mime-edit-maybe-translate)
      (mime-edit-translate-buffer))
    (message (format "Message converted to %d-bit MIME." bit-level))))

(defun cz-message-convert-to-csech ()
  "Convert message to ASCII."
  (goto-char (point-min))
  (if (re-search-forward mail-header-separator nil t)
      (progn
	(beginning-of-line)
	(cz-message-convert-to-csech-region (point-min) (point))
	(forward-line)
	(cz-message-convert-to-csech-region (point) (point-max)))
    (cz-message-convert-to-csech-region (point-min) (point-max))))

(defun cz-message-convert-to-csech-region (beg end)
  "Convert region to ASCII."
  ;; "Manual" conversion
  ;; After Emacs 20 compliant ogonek.el will be available, we will use it here
  (save-restriction
    (narrow-to-region beg end)
    (goto-char (point-min))
    (let ((string "[,B !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~(B]")
	  (table "[ A L LS \"SSTZ-ZZ a l'ls  sstz\"zzRAAAALCCCEEEEIIDDNNOOOO.RUUUUYT raaaalccceeeeiiddnnoooo/ruuuuyt ]")
	  char
	  (case-fold-search nil))
      (while (re-search-forward string nil t)
	(setq char (aref table (string-match (char-to-string (preceding-char))
					     string)))
	(delete-char -1)
	(insert char)))))

(defun cz-message-convert-to-digraphs ()
  "Convert subject and body of the message to enchainments encoding."
  (require 'string-recode)
  (save-excursion
    (goto-char (point-min))
    (re-search-forward "^Subject:" nil t)
    (sr-recode-to-many
     (point)
     (progn (re-search-forward "^[^ \t]") (forward-char -2) (point))
     "czech-digraphs")
    (sr-recode-to-many
     (progn
       (goto-char (point-min))
       (or (search-forward mail-header-separator nil t)
	   (point-max)))
     (point-max)
     "czech-digraphs")))
    
(defun cz-message-method ()
  "Find conversion method of the current message.
Method is found according to values of miscellaneous variables, especially
`cz-gnus', `cz-message-method-default', `cz-message-method-people', and
`cz-message-news'."
  (if (and (not (cz-message-rmail-p))
	   (message-news-p))	      ; if both news and mail, news is prefered
      (save-excursion
	(save-restriction
	  (message-narrow-to-headers)
	  (cz-message-find-method
	   (cz-message-rfc822-addresses "newsgroups" "followup-to")
	   cz-message-method-newsgroups
	   cz-message-method-news
	   nil)))
    (save-excursion
      (save-restriction
	(message-narrow-to-headers)
	(cz-message-find-method
	 (cz-message-rfc822-addresses "to" "cc" "bcc")
	 cz-message-method-people
	 cz-message-method-mail
	 cz-message-use-bbdb)))))

(defun cz-message-rfc822-addresses (&rest fields)
  "Auxiliary for getting addresses even containing 8-bit characters."
  (let ((orig (buffer-substring (point-min) (point-max)))
	string
	addresses)
    ;; Unaccent
    (cz-message-convert-to-csech)
    ;; Find addresses
    (while fields
      (setq string (concat string (message-fetch-field (car fields)) ","))
      (setq fields (cdr fields)))
    (setq addresses (rfc822-addresses string))
    ;; Return accents
    (delete-region (point-min) (point-max))
    (insert orig)
    addresses))

(defun cz-message-find-method (recipients recip-method-list default use-bbdb)
  "Find conversion method of message according to given variables.
Returns one of the following symbols: `ascii', `mime', `mime-8', `plain'.
If USE-BBDB is non-nil, use Insidious Big Brother Database
`mailtype' (user defined) field."
  (let (methods)
    (while recipients
      (setq methods (cons (cz-message-find-method-one (car recipients)
						      recip-method-list
						      default
						      use-bbdb)
			  methods))
      (setq recipients (cdr recipients)))
    ;; Select "the weakest" method
    (cond
     ((memq 'ascii methods)
      'ascii)
     ((memq 'mime methods)
      (if (or (memq 'plain methods) (memq 'digraphs methods))
	      'ascii
	    'mime))
     ((memq 'digraphs methods)
      (if (memq 'plain methods)
	  'ascii
	'digraphs))
     ((memq 'mime-8 methods)
      'mime-8)
     (t
      'plain))))

(defun cz-message-find-method-one (recipient recip-method-list default
					     use-bbdb)
  (or
   ;; Try to find in bbdb
   (and use-bbdb
	(or (featurep 'bbdb) (load "bbdb" t))
	(cz-message-find-bbdb recipient))
   ;; Try to find in the list
   (cz-message-find-list recipient recip-method-list)
   ;; Find default value
   (progn
     (if (not default)
	 (setq default cz-message-method-default))
     (if (eq default 'plain)
	 nil
       default))))

(defun cz-message-find-list (recipient list)
  (let (method
	aux)
    (while (and (not method) list)
      (setq aux (cdr (car list)))	; skip type
      (while (and (not method) aux)
	(if (string-match (car aux) recipient)
	    (setq method (car (car list)))
	  (setq aux (cdr aux))))
      (setq list (cdr list)))
    method))

(defun cz-message-find-bbdb (recipient)
  (let ((record (bbdb-search-simple nil recipient))
	(aux))
    (and record
	 (setq aux (bbdb-record-getprop record 'mailtype))
	 (intern aux))))


;;; *** Miscellaneous***

(defvar cz-message-history nil
  "History for `cz-message-set-method'.")

;;;###autoload
(defun cz-message-set-method ()
  "Sets final encoding of the current message."
  (interactive)
  (let ((method (completing-read
		 "Message encoding: "
		 (if (eq cz-gnus t)
		     '(("mime" 1) ("ascii" 2) ("plain" 3) ("sprzezjky" 4)
		       ("default" 5))
		   '(("ascii" 2) ("plain" 3) ("sprzezjky" 4) ("default" 5)))
		 nil nil nil
		 'cz-message-history)))
    (setq cz-message-this-method
	  (if (string= method "sprzezjky")
	      'digraphs
	    (intern method)))))

;;;###autoload
(defun cz-message-show-method (&optional quietly)
  "Prints final encoding of the current message.

If an optional argument QUIETLY is non-nil, no message is printed, only
symbol value is returned."
  (interactive)
  (let ((method (if (not (eq cz-message-this-method 'default))
		    cz-message-this-method
		  (cz-message-method))))
    ;; Check MIME availability
    (if (and (or (eq method 'mime) (eq method 'mime-8))
	     (not (eq cz-gnus t)))
	(progn
	  (if quietly
	      (message "MIME method requested but forbidden."))
	  (setq method 'ascii)))
    ;; If `plain', use 8-bit MIME if available
    (if (and (eq method 'plain) cz-message-evil-sysadmin)
	(setq method (if (eq cz-gnus t) 'mime-8 nil)))
    ;; Final work around
    (if quietly
	method
      (message (format "On the current message would be aplied method %s."
		       (cond
			((eq method 'ascii) "ASCII")
			((eq method 'mime) "MIME (7-bit)")
			((eq method 'mime-8) "MIME (8-bit)")
			((eq method 'digraphs) "sprzezjky")
			((or (eq method 'plain) (null method)) "plain")))))))


;;; *** Announce ***

(provide 'czech-message)


;;; czech-message.el ends here

