;;; string-recode.el - recoding of sequences

;; Copyright (C) 1997, 1998 Milan Zamazal

;; Author:     Milan Zamazal <pdm@fi.muni.cz>
;; Version:    4.3
;; Keywords:   encodings, conversion, Czech

;; COPYRIGHT NOTICE
;;
;; This program is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the Free
;; Software Foundation, version 2 of the License.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
;; for more details.
;;
;; You can find the GNU General Public License at
;; http://www.gnu.org/copyleft/gpl.html
;; or you can write to the Free Software Foundation, Inc., 675 Mass Ave.,
;; Cambridge, MA 02139, USA.

;;; Installation:

;; Put this file somewhere into load-path and to your ~/.emacs the following
;; lines:
;;
;;   (autoload 'sr-recode-to-many "string-recode" t)
;;   (autoload 'sr-recode-from-many "string-recode" t)
;;
;; It may also be a good idea to append:
;;
;;   (add-hook 'sgml-mode-hook (lambda () (setq sr-default-table "SGML")))

;;; Commentary:

;; This program performs recoding between strings in the text.  Its primary
;; usage is conversion between Czech accented characters and appropriate
;; (La)TeX sequences.
;;
;; Through the variable `sr-tables' you can specify one input string and one or
;; more output strings.  If more than one output string is specified, then for
;; input-output conversion the first one is used.  For the output-input
;; conversion, the most often appearing sequence in the corrected text is
;; supposed to be the right one and this is the only converted for each given
;; input string.
;;
;; Currently supported conversions are TeX sequences, SGML sequences, and
;; digraphs (the latter is mostly joke :-).

;; This code is probably obsoleted by the CCL.
;; If you can define recodings using this system, please do so and send me the
;; result for inclusion into the next version of emacs-cs-utils.

;;; Code:


(defvar sr-default-table "TeX"
  "*Default recoding table.
See `sr-tables'.")
(make-variable-buffer-local 'sr-default-table)

(defvar sr-tables
  '(("TeX" ""
     ("" "\\'a" "\\'{a}")
     ("" "\\\"a" "\\\"{a}")
     ("" "\\v c" "\\v{c}")
     ("" "\\v d" "\\v{d}")
     ("" "\\'e" "\\'{e}")
     ("" "\\v e" "\\v{e}")
     ("" "\\'\\i{}" "\\'{\\i}")
     ("" "\\'l" "\\'{l}")
     ("" "\\v l" "\\v{l}")
     ("" "\\v n" "\\v{n}")
     ("" "\\'o" "\\{'o}")
     ("" "\\\"o" "\\\"{o}")
     ("" "\\^o" "\\^{o}")
     ("" "\\'r" "\\'{r}")
     ("" "\\v r" "\\v{r}")
     ("" "\\v s" "\\v{s}")
     ("" "\\v t" "\\v{t}")
     ("" "\\'u" "\\'{u}")
     ("" "\\accent23u" "\\r{u}")
     ("" "\\\"u" "\\\"{u}")
     ("" "\\'y" "\\'{y}")
     ("" "\\v z" "\\v{z}")
     ("" "\\'A" "\\'{A}")
     ("" "\\\"A" "\\\"{A}")
     ("" "\\v C" "\\v{C}")
     ("" "\\v D" "\\v{D}")
     ("" "\\'E" "\\'{E}")
     ("" "\\v E" "\\v{E}")
     ("" "\\'\\I{}" "\\'{\\I}")
     ("" "\\'L" "\\'{L}")
     ("" "\\v L" "\\v{L}")
     ("" "\\v N" "\\v{N}")
     ("" "\\'O" "\\{'O}")
     ("" "\\\"O" "\\\"{O}")
     ("" "\\^O" "\\^{O}")
     ("" "\\'R" "\\'{R}")
     ("" "\\v R" "\\v{R}")
     ("" "\\v S" "\\v{S}")
     ("" "\\v T" "\\v{T}")
     ("" "\\'U" "\\'{U}")
     ("" "\\accent23U" "\\r{U}")
     ("" "\\\"U" "\\\"{U}")
     ("" "\\'Y" "\\'{Y}")
     ("" "\\v Z" "\\v{Z}"))
    ("SGML" ""
     ("" "&aacute;")
     ("" "&auml;")
     ("" "&ccaron;")
     ("" "&dcaron;")
     ("" "&eacute;")
     ("" "&ecaron;")
     ("" "&iacute;")
     ("" "&lacute;")
     ("" "&lcaron;")
     ("" "&ncaron;")
     ("" "&oacute;")
     ("" "&ouml;")
     ("" "&ocirc;")
     ("" "&rcaron;")
     ("" "&racute;")
     ("" "&scaron;")
     ("" "&tcaron;")
     ("" "&ucaron;")
     ("" "&uring;")
     ("" "&uuml;")
     ("" "&yacute;")
     ("" "&zcaron;")
     ("" "&Aacute;")
     ("" "&Auml;")
     ("" "&Ccaron;")
     ("" "&Dcaron;")
     ("" "&Eacute;")
     ("" "&Ecaron;")
     ("" "&Iacute;")
     ("" "&Lacute;")
     ("" "&Lcaron;")
     ("" "&Ncaron;")
     ("" "&Oacute;")
     ("" "&Ouml;")
     ("" "&Ocirc;")
     ("" "&Rcaron;")
     ("" "&Racute;")
     ("" "&Scaron;")
     ("" "&Tcaron;")
     ("" "&Ucaron;")
     ("" "&Uring;")
     ("" "&Uuml;")
     ("" "&Yacute;")
     ("" "&Zcaron;"))
    ("czech-digraphs" "|"
     ("d" "dje")
     ("t" "tje")
     ("n" "nje")
     ("b" "bje")
     ("p" "pje")
     ("v" "vje")
     ("m" "mje")
     (""  "aa")
     (""  "ee")
     (""  "je")
     (""  "ii")
     (""  "oo")
     (""  "uu")
     (""  "uo" "uu")
     (""  "yy")
     (""  "cz" "cj")
     (""  "rz" "rj")
     (""  "sz" "sj")
     (""  "dj")
     (""  "nj")
     (""  "tj")
     (""  "zj" "zh" "zz")
     ("D" "Dje")
     ("T" "Tje")
     ("N" "Nje")
     ("B" "Bje")
     ("P" "Pje")
     ("V" "Vje")
     ("M" "Mje")
     (""  "Aa")
     (""  "Ee")
     (""  "Ii")
     (""  "Oo")
     (""  "Uu")
     (""  "Uo" "Uu")
     (""  "Yy")
     (""  "Cz" "Cj")
     (""  "Rz" "Rj")
     (""  "Sz" "Sj")
     (""  "Dj")
     (""  "Nj")
     (""  "Tj")
     (""  "Zj" "Zh" "Zz")
     ("D" "DJE")
     ("T" "TJE")
     ("N" "NJE")
     ("B" "BJE")
     ("P" "PJE")
     ("V" "VJE")
     ("M" "MJE")
     (""  "AA")
     (""  "EE")
     (""  "II")
     (""  "OO")
     (""  "UU")
     (""  "UO" "UU")
     (""  "YY")
     (""  "CZ" "CJ")
     (""  "RZ" "RJ")
     (""  "SZ" "SJ")
     (""  "DJ")
     (""  "NJ")
     (""  "TJ")
     (""  "ZJ" "ZH" "ZZ")))
  "Recoding tables.

It is an association list which elements are of the form

  (NAME SEPARATOR (PATTERN ENCODED ...) ...)

NAME is a name of the recoding.
SEPARATOR is a string inserted when some original text could be confused with
an encoded text.
PATTERN is a string to be encoded.
The first ENCODED element is the encoded PATTERN.  Optional next ENCODED
elements are non-canonical encodings of the PATTERN used in back conversion.")

;;;###autoload
(defun sr-recode-from-many (beg end name)
  "Recode region to the first elements the item NAME of the `sr-tables'."
  (interactive
   (progn
     (barf-if-buffer-read-only)
     (list
      (region-beginning)
      (region-end)
      (sr-read-name))))
  (let* ((end-marker (copy-marker end))
	 (item (cdr (assoc name sr-tables)))
	 (separator (car item))
	 (table (cdr item))
	 recoding-list
	 regexp)
    (save-excursion
      ;; Build replacement list
      (let (winner
	    winner-count
	    choices
	    count)
	(while table
	  (setq winner (car (cdar table))
		winner-count (sr-regexp-count winner beg end)
		choices (cdr (cdar table)))
	  ;; We choose the most often repeating encoding element as the right
	  ;; one used by the user.
	  (while choices
	    (setq count (sr-regexp-count (car choices) beg end))
	    (if (> count winner-count)
		(setq winner (car choices)
		      winner-count count))
	    (setq choices (cdr choices)))
	  (setq recoding-list (cons (cons winner (caar table))
				    recoding-list)
		regexp (if (null regexp)
			   (regexp-quote winner)
			 (concat regexp "\\|" (regexp-quote winner))))
	  (setq table (cdr table))))
      ;; Recode strings
      (goto-char beg)
      (let ((case-fold-search nil))
	(while (and (<= (point) (marker-position end-marker))
		    (re-search-forward regexp end-marker t))
	  (replace-match (cdr (assoc (match-string 0) recoding-list)) 1)))
      ;; Remove separators
      (unless (string= separator "")
	(goto-char beg)
	(while (and (<= (point) (marker-position end-marker))
		    (search-forward separator end-marker t))
	  (delete-char -1)
	  (forward-char 1))))))

;;;###autoload
(defun sr-recode-to-many (beg end name)
  "Recode region to the second elements of the NAME item of `sr-tables'."
  (interactive
   (progn
     (barf-if-buffer-read-only)
     (list
      (region-beginning) 
      (region-end)
      (sr-read-name))))
  (let* ((end-marker (copy-marker end))
	 (item (cdr (assoc name sr-tables)))
	 (separator (car item))
	 (table (cdr item))
	 (regexp (concat (caar table) "\\|" (car (cdar table)))))
    (save-excursion
      ;; Build regular expression
      (let ((table-copy (cdr table))
	    aux)
	(while table-copy
	  (setq aux (car table-copy))
	  (setq regexp (concat regexp "\\|" (car aux) "\\|" (cadr aux)))
	  (setq table-copy (cdr table-copy))))
      ;; Start searching
      (goto-char beg)
      (let ((case-fold-search nil)
	    string-item)
	(while (and (<= (point) (marker-position end-marker))
		    (re-search-forward regexp end-marker t))
	  (setq string-item (assoc (match-string 0) table))
	  (if string-item
	      ;; Replace by an appropriately encoded string
	      (replace-match (regexp-quote (cadr string-item)) t)
	    ;; Insert separator
	    (goto-char (1+ (match-beginning 0)))
	    (insert separator)))))))


;;; Auxiliary functions

(defvar sr-names-history nil)

(defun sr-read-name ()
  "Read name of conversion from minibuffer."
  (setq sr-default-table
	(completing-read
	 (format "Conversion type (default %s): " sr-default-table)
	 (let ((tables sr-tables)
	       choices)
	   (while tables
	     (setq choices (cons (list (caar tables)) choices))
	     (setq tables (cdr tables)))
	   choices)
	 nil t "" 'sr-names-history sr-default-table)))

(defun sr-regexp-count (regexp beg end)
  "Number of REGEXP occurences in the region."
  (goto-char beg)
  (let ((n 0))
    (while (search-forward regexp end t)
      (setq n (1+ n)))
    n))


;;; Announce

(provide 'string-recode)


;; Local variables:
;; coding: iso-latin-2
;; End:

;;; string-recode.el ends here
